import traceback

import numpy as np
import logging
import onnxruntime
import pickle
import requests
from io import BytesIO
from PIL import Image

from social_net_img_classifier.aws_utils import AwsS3
from social_net_img_classifier.settings import MODEL_NUDE_DIR

if Image is not None:
    _PIL_INTERPOLATION_METHODS = {
        "nearest": Image.NEAREST,
        "bilinear": Image.BILINEAR,
        "bicubic": Image.BICUBIC,
    }
    # These methods were only introduced in version 3.4.0 (2016).
    if hasattr(Image, "HAMMING"):
        _PIL_INTERPOLATION_METHODS["hamming"] = Image.HAMMING
    if hasattr(Image, "BOX"):
        _PIL_INTERPOLATION_METHODS["box"] = Image.BOX
    # This method is new in version 1.1.3 (2013).
    if hasattr(Image, "LANCZOS"):
        _PIL_INTERPOLATION_METHODS["lanczos"] = Image.LANCZOS


class NudeClassifier:
    """
    Class for loading model and running predictions.
    For example on how to use take a look the if __name__ == '__main__' part.
    """

    nsfw_model = None

    def __init__(self):
        """
        model = Classifier()
        """
        model_bytes = pickle.loads(AwsS3().read_file(MODEL_NUDE_DIR))
        self.nsfw_model = onnxruntime.InferenceSession(model_bytes)

    def classify(
        self,
        image_paths=[],
        batch_size=4,
        image_size=(256, 256),
        categories=["unsafe", "safe"],
    ):
        """
        inputs:
            image_paths: list of image paths or can be a string too (for single image)
            batch_size: batch_size for running predictions
            image_size: size to which the image needs to be resized
            categories: since the model predicts numbers, categories is the list of actual names of categories
        """
        try:
            if not isinstance(image_paths, list):
                image_paths = [image_paths]

            loaded_images, loaded_image_paths = load_images(
                image_paths, image_size, image_names=image_paths
            )

            if not loaded_image_paths:
                return {}

            preds = []
            model_preds = []
            while len(loaded_images):
                _model_preds = self.nsfw_model.run(
                    [self.nsfw_model.get_outputs()[0].name],
                    {self.nsfw_model.get_inputs()[0].name: loaded_images[:batch_size]},
                )[0]
                model_preds.append(_model_preds)
                preds += np.argsort(_model_preds, axis=1).tolist()
                loaded_images = loaded_images[batch_size:]

            probs = []
            for i, single_preds in enumerate(preds):
                single_probs = []
                for j, pred in enumerate(single_preds):
                    single_probs.append(
                        model_preds[int(i / batch_size)][int(i % batch_size)][pred]
                    )
                    preds[i][j] = categories[pred]

                probs.append(single_probs)

            images_preds = {}

            for i, loaded_image_path in enumerate(loaded_image_paths):
                if not isinstance(loaded_image_path, str):
                    loaded_image_path = i

                images_preds[loaded_image_path] = {}
                for _ in range(len(preds[i])):
                    images_preds[loaded_image_path][preds[i][_]] = float(probs[i][_])
            return images_preds
        except Exception:
            traceback.print_exc()
            return {}


def load_images(image_paths, image_size, image_names):
    """
    Function for loading images into numpy arrays for passing to model.predict
    inputs:
        image_paths: list of image paths to load
        image_size: size into which images should be resized

    outputs:
        loaded_images: loaded images on which keras model can run predictions
        loaded_image_indexes: paths of images which the function is able to process

    """
    loaded_images = []
    loaded_image_paths = []

    for i, img_path in enumerate(image_paths):
        try:
            image = load_img(img_path, target_size=image_size)
            image = img_to_array(image)
            image /= 255
            loaded_images.append(image)
            loaded_image_paths.append(image_names[i])
        except Exception as ex:
            logging.exception(f"Error reading {img_path} {ex}", exc_info=True)

    return np.asarray(loaded_images), loaded_image_paths


def load_img(
        path, grayscale=False, color_mode="rgb", target_size=None, interpolation="nearest"
):
    """Loads an image into PIL format.

    :param path: Path to image file.
    :param grayscale: DEPRECATED use `color_mode="grayscale"`.
    :param color_mode: One of "grayscale", "rgb", "rgba". Default: "rgb".
        The desired image format.
    :param target_size: Either `None` (default to original size)
        or tuple of ints `(img_height, img_width)`.
    :param interpolation: Interpolation method used to resample the image if the
        target size is different from that of the loaded image.
        Supported methods are "nearest", "bilinear", and "bicubic".
        If PIL version 1.1.3 or newer is installed, "lanczos" is also
        supported. If PIL version 3.4.0 or newer is installed, "box" and
        "hamming" are also supported. By default, "nearest" is used.

    :return: A PIL Image instance.
    """
    if grayscale is True:
        logging.warn("grayscale is deprecated. Please use " 'color_mode = "grayscale"')
        color_mode = "grayscale"
    if Image is None:
        raise ImportError(
            "Could not import PIL.Image. " "The use of `load_img` requires PIL."
        )

    if path[:5] == "http:" or path[:6] == "https:":
        response = requests.get(path)
        img = Image.open(BytesIO(response.content))
    else:
        img = Image.open(path)

    if color_mode == "grayscale":
        if img.mode != "L":
            img = img.convert("L")
    elif color_mode == "rgba":
        if img.mode != "RGBA":
            img = img.convert("RGBA")
    elif color_mode == "rgb":
        if img.mode != "RGB":
            img = img.convert("RGB")
    else:
        raise ValueError('color_mode must be "grayscale", "rgb", or "rgba"')
    if target_size is not None:
        width_height_tuple = (target_size[1], target_size[0])
        if img.size != width_height_tuple:
            if interpolation not in _PIL_INTERPOLATION_METHODS:
                raise ValueError(
                    "Invalid interpolation method {} specified. Supported "
                    "methods are {}".format(
                        interpolation, ", ".join(_PIL_INTERPOLATION_METHODS.keys())
                    )
                )
            resample = _PIL_INTERPOLATION_METHODS[interpolation]
            img = img.resize(width_height_tuple, resample)
    return img


def img_to_array(img, data_format="channels_last", dtype="float32"):
    """Converts a PIL Image instance to a Numpy array.
    # Arguments
        img: PIL Image instance.
        data_format: Image data format,
            either "channels_first" or "channels_last".
        dtype: Dtype to use for the returned array.
    # Returns
        A 3D Numpy array.
    # Raises
        ValueError: if invalid `img` or `data_format` is passed.
    """
    if data_format not in {"channels_first", "channels_last"}:
        raise ValueError("Unknown data_format: %s" % data_format)
    # Numpy array x has format (height, width, channel)
    # or (channel, height, width)
    # but original PIL image has format (width, height, channel)
    x = np.asarray(img, dtype=dtype)
    if len(x.shape) == 3:
        if data_format == "channels_first":
            x = x.transpose(2, 0, 1)
    elif len(x.shape) == 2:
        if data_format == "channels_first":
            x = x.reshape((1, x.shape[0], x.shape[1]))
        else:
            x = x.reshape((x.shape[0], x.shape[1], 1))
    else:
        raise ValueError("Unsupported image shape: %s" % (x.shape,))
    return x
